%{
/* C declarations */

typedef struct yystype;
{
    int foo;
} yystype;
#define YYSTYPE yystype;

%}
/* BISON declarations */

/* Operator precedence - least precedence group */
%left IMP EQV   /* 1101, 1001 */
%left XOR OR    /* 0110, 0111 */
%left AND       /* 0001 */

%right NOT  /* 0=1 1=0 */

/* Operator precedence - middle precedence group */
%left '<' '>' '=' LESSOREQUAL GREATEROREQUAL NOTEQUAL

/* Operator precedence - top precedence group */
%left '+' '-'
%left MOD
%left '\\'
%left '*' '/'

%right NEGATIONPRIORITY
%right '^'

%token CRLF ':'

%token IF THEN ELSE ELSEIF ENDIF

%token SELECT CASE ENDSELECT IS

%token WEND DO LOOP UNTIL WHILE EXITDO

%token SUB ENDSUB EXITSUB FUNCTION ENDFUNCTION EXITFUNCTION

%token FOR TO STEP NEXT EXITFOR
%token GOTO GOSUB RETURN
%token END
%token IDENTIFIER NUMBER STRING
%token LET

%token PRINT USING FILENUM LPRINT

%%
/* Grammar rules */

program:
		definitionlist
;

definitionlist:
		definitionlist definition
	|	/* empty */
;

definition:
		subdeclaration
	|	functiondeclaration
	|	longstatement
	|	CRLF
;

subdeclaration:
		SUB var CRLF longstatements ENDSUB CRLF
;
functiondeclaration:
		FUNCTION var CRLF longstatements ENDFUNCTION CRLF
;

/* This does not terminate with anything */
shortstatements:
		shortstatementsnotempty
	|	/* empty */
;
shortstatementsnotempty:
		shortstatement
	|	shortstatementsnotempty ':' shortstatement
;

/* This does not terminate with anything */
shortstatement:
		shortstructure
	|	simplestatement
;

/* This does not terminate with anything */
shortstructure:
		IF boolexpr THEN shortstatements shortelselist
	|	WHILE boolexpr ':' shortstatementsnotempty WEND
	|	WHILE boolexpr ':' WEND
	|	DO doexpr ':' shortstatementsnotempty LOOP doexpr
	|	DO doexpr ':' LOOP doexpr
	|	FOR var '=' numexpr TO numexpr forstep ':' shortstatementsnotempty NEXT
	|	FOR var '=' numexpr TO numexpr forstep ':' NEXT
;

forstep:
		STEP numexpr
	|	/* empty */
;

shortelseiflist:
		shortelseiflist ELSEIF boolexpr THEN shortstatements
	|	/* empty */
;
shortelselist:
		shortelseiflist ELSE shortstatements
	|	shortelseiflist
;

longstatements:
		longstatementsnotempty
	|	/* empty */
;
longstatementsnotempty:
		longstatement separator
	|	longstatementsnotempty longstatement separator
;
longstatement:
		longstructure
	|	shortstatementsnotempty
;
longstructure:
		IF boolexpr THEN CRLF longstatements longelselist ENDIF
	|	WHILE boolexpr separator longstatements WEND
	|	DO doexpr separator longstatements LOOP doexpr
	|	FOR var '=' numexpr TO numexpr forstep separator longstatements NEXT
	|	SELECT CASE expr separator caselist ENDSELECT
;

casevalue:
		ELSE
	|	IS '>' expr
	|	IS '<' expr
	|	IS '=' expr
	|	IS LESSOREQUAL expr
	|	IS GREATEROREQUAL expr
	|	IS NOTEQUAL expr
	|	expr TO expr
	|	expr
;
casevalues:
	|	casevalues ',' casevalue
	|	casevalue
;

caselist:
		caselist CASE casevalues separator longstatements
	|	/* empty */
;

separator:
		':'
	|	newline
;

newline:
	|	CRLF NUMBER
	|	CRLF label ':'
	|	CRLF /* shift-reduce conflict, I know */
;

longelselist:
		ELSEIF boolexpr THEN separator longstatements longelselist
	|	ELSE separator longstatements
	|	/* empty */
;

doexpr:
		WHILE boolexpr
	|	UNTIL boolexpr
	|	/* empty */
;

/* Something that is not a control structure */
/* This does not terminate with anything */
simplestatement:
		LET assignment
	|	assignment
	|	GOTO label
	|	GOSUB label
	|	GOTO NUMBER
	|	GOSUB NUMBER
	|	RETURN
	|	EXITDO
	|	EXITFOR
	|	EXITSUB
	|	EXITFUNCTION
	|	print
	|	END
;

assignment:
		var '=' expr
;

print:
		PRINT FILENUM ',' printspec
	|	PRINT printspec
	|	LPRINT printspec
;
printspec:
		USING stringexpr ';' printparams
	|	printparams
;
printparams:
		printparams ';' expr  /* list; expr */
	|	printparams ',' expr  /* list, expr */
	|	printparams ';'       /* list;  */
	|	printparams ','       /* list,  */
	|	expr                  /* expr */
	|	/* empty */
;

numexpr:
		expr	/* FIXME - check expression type */
;
boolexpr:
		expr	/* FIXME - check expression type */
;
stringexpr:
		expr	/* FIXME - check expression type */
;

expr:
		expr IMP expr
	|	expr EQV expr
	|	expr XOR expr
	|	expr OR expr
	|	expr AND expr
	|	NOT expr
	
	|	expr '>' expr
	|	expr '<' expr
	|	expr '=' expr
	|	expr LESSOREQUAL expr
	|	expr GREATEROREQUAL expr
	|	expr NOTEQUAL expr

	|	expr '+' expr
	|	expr '-' expr
	|	expr MOD expr
	|	expr '\\' expr
	|	expr '/' expr
	|	expr '*' expr
	|	expr '^'expr
	
	|       '-' expr %prec NEGATIONPRIORITY

	|	'(' expr ')'
	|	var
	|	NUMBER
	|	STRING
;

indexlist:
		indexlist ',' numexpr
	|	numexpr
;
label:
		IDENTIFIER
;
var:
		IDENTIFIER
	|	IDENTIFIER '(' indexlist ')' /* array indexing */
;

%%
/* Additional C code */
